import os
import sys
import pyfits

sys.path.append("/utils")
import utils

CONFIG = os.environ["CONFIG"]
INDIR = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

params = cab["parameters"]

_positional = ['antenna-file']
positional = []
for item in _positional:
    param = filter( lambda a: a['name']==item, params)[0]
    if param['value']:
        positional.append(param['value'])

#f params.pop("from-fits", False) and imagename:
#   with pyfits.open(imagename) as hdu:
#       hdr = hdu[0].header
#       naxis = hdr["naxis"]

#       freq = filter(lambda ind: hdr["ctype%d"%ind].startswith("FREQ"),
#               range(2, naxis+1, 1))
#       if freq:
#           freq = freq[0]
#       else:
#           raise TypeError("Your FITS image has no frequency information")

#       nchan = hdr["naxis%d"%freq]
#       freq0 = hdr["crval%d"%freq]
#       dfreq = abs(hdr["cdelt%d"%freq])

#       params["freq0"] = freq0
#       params["dfreq"] = dfreq
#       params["nchan"] = nchan

#       if not direction:
#           params["direction"] = "J2000,%fdeg,%fdeg"%(hdr["crval1"], hdr["crval2"])

args = []
for param in params:
    key = param['name']
    value = param['value']
    if value in [None, False]:
        continue
    if value is True:
        arg = '{0}{1}'.format(cab['prefix'], key)
    elif hasattr(value, '__iter__'):
        arg = ' '.join(['{0}{1} {2}'.format(cab['prefix'], key, val) for val in value])
    else:
        if isinstance(value, str):
            value = '"{}"'.format(value)
        arg = '{0}{1} {2}'.format(cab['prefix'], key, value)

    args.append(arg)

utils.xrun(cab['binary'], args+['--nolog']+positional)



import os
import sys
import time
import stimela
from stimela import docker, utils, cargo
from stimela.cargo import cab
import logging
import inspect
import re

USER = os.environ["USER"]
UID = os.getuid()
CAB_PATH = os.path.abspath(os.path.dirname(cab.__file__))


class PipelineException(Exception):
    """ 
    Encapsulates information about state of pipeline when an
    exception occurs
    """
    def __init__(self, exception, completed, failed, remaining):
        message = ("Exception occurred while running "
            "pipeline component '%s': %s" % (failed.label, str(exception)))

        super(PipelineException, self).__init__(message)

        self._completed = completed
        self._failed = failed
        self._remaining = remaining

    @property
    def completed(self):
        return self._completed

    @property
    def failed(self):
        return self._failed

    @property
    def remaining(self):
        return self._remaining

class Recipe(object):
    def __init__(self, name, data=None,
                 parameter_file_dir=None, ms_dir=None,
                 tag=None, loglevel='INFO'):
        """
        Deifine and manage a stimela recipe instance.        

        name    :   Name of stimela recipe
        data    :   Path of stimela data. The data is assumed to be at Stimela/stimela/cargo/data
        msdir   :   Path of MSs to be used during the execution of the recipe
        tag     :   Use cabs with a specific tag
        parameter_file_dir :   Will store task specific parameter files here
        """

        self.log = logging.getLogger('STIMELA')
        self.log.setLevel(getattr(logging, loglevel))
        # create file handler which logs even debug
        # messages
        name_ = name.lower().replace(' ', '_')
        self.logfile = 'log-{}.txt'.format(name_)
        self.resume_file = '.last_{}.json'.format(name_)

        fh = logging.FileHandler(self.logfile)
        fh.setLevel(logging.DEBUG)
        # create console handler with a higher log level
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.ERROR)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        fh.setFormatter(formatter)
        # add the handlers to logger
        self.log.addHandler(ch)
        self.log.addHandler(fh)

        self.stimela_context = inspect.currentframe().f_back.f_globals

        self.stimela_path = os.path.dirname(docker.__file__)

        self.name = name
        self.ms_dir = ms_dir
        if not os.path.exists(self.ms_dir):
            self.log.info('MS directory \'{}\' does not exist. Will create it'.format(self.ms_dir))
            os.mkdir(self.ms_dir)
        self.tag = tag
        # create a folder to store config files
        # if it doesn't exist. These config
        # files can be resued to re-run the
        # task
        self.parameter_file_dir = parameter_file_dir or "stimela_parameter_files"
        if not os.path.exists(self.parameter_file_dir):
            self.log.info('Config directory not be found. Will create ./{}'.format(self.parameter_file_dir))
            os.mkdir(self.parameter_file_dir)

        self.containers = []
        self.completed = []
        self.failed = None
        self.remaining = []

        self.proc_logger = utils.logger.StimelaLogger(stimela.LOG_FILE)
        self.pid = os.getpid()
        self.proc_logger.log_process(self.pid, self.name)
        self.proc_logger.write()


    def add(self, image, name=None, config=None,
            input=None, output=None, msdir=None,
            label=None, shared_memory='1gb'):
        """
        Add a task to a stimela recipe

        image   :   stimela cab name, e.g. 'cab/simms'
        name    :   This name will be part of the name of the contaier that will 
                    execute the task (now optional)
        config  :   Dictionary of options to parse to the task. This will modify 
                    the parameters in the default parameter file which 
                    can be viewd by running 'stimela cabs -i <cab name>', e.g 'stimela cabs -i simms'
        input   :   input dirctory for cab
        output  :   output directory for cab
        msdir   :   MS directory for cab. Only specify if different from recipe ms_dir
        """

        # check if name has any offending charecters
        offenders = re.findall('\W', name)
        if offenders:
            raise ValueError('The cab name \'{:s}\' has some non-alphanumeric characters.'
                             ' Charecters making up this name must be in [a-z,A-Z,0-9,_]'.format(name))
        # Get location of template parameters file
        parameter_file = '{0}/{1}/parameters.json'.format(CAB_PATH, image.split('/')[-1].split(':')[0])
        msdir = msdir or self.ms_dir
        
        if name is None:
            name = '{0}_{1}-{2}{3}'.format(USER, image, id(image), str(time.time()).replace('.', ''))
        else:
            name = '{0}-{1}{2}'.format(name, id(image), str(time.time()).replace('.', ''))

        _cab = cab.CabDefinition(indir=input, outdir=output,
                    msdir=msdir, parameter_file=parameter_file)
         


        # Volumes to be mounted into the container
        input = input or self.stimela_context.get('STIMELA_INPUT', None)
        output = output or self.stimela_context.get('STIMELA_OUTPUT', None)

        cont = docker.Container(image, name,
                     label=label, logger=self.log,
                     shared_memory=shared_memory, log_container=stimela.LOG_FILE)

        
#        parameter_file_name = '{0}/{1}.json'.format(self.parameter_file_dir, name)
#        _cab.update(config, parameter_file_name)
        
        # Container parameter file will be updated and validated before the container is executed
        cont._cab = _cab
        cont.parameter_file_name = '{0}/{1}.json'.format(self.parameter_file_dir, name)
        cont.config = config

        # These are standard volumes and
        # environmental variables. These will be
        # always exist in a cab container
        cont.add_volume(self.stimela_path, '/utils', perm='ro')
        cont.add_volume(self.parameter_file_dir, '/configs', perm='ro')
        cont.add_environ('CONFIG', '/configs/{}.json'.format(name))

        if msdir:
            md = '/home/%s/msdir'%USER
            cont.add_volume(msdir, md)
            cont.add_environ('MSDIR', md)
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(msdir))]
            cont.msdir_content = {
                "volume"    :   dirname,
                "dirs"      :   dirs,
                "files"     :   files,
            }

            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(msdir, md))

        if input:
            cont.add_volume( input,'/input', perm='ro')
            cont.add_environ('INPUT', '/input')
            # Keep a record of the content of the
            # volume
            dirname, dirs, files = [a for a in next(os.walk(input))]
            cont.input_content = {
                "volume"    :   dirname,
                "dirs"      :   dirs,
                "files"     :   files,
            }

            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(input, '/input'))

        if output:
            if not os.path.exists(output):
                os.mkdir(output)
            od = '/home/%s/output'%USER
            cont.add_volume(output, od)
            cont.add_environ('OUTPUT', od)
            cont.logfile = '{0}/log-{1}.txt'.format(output, name.split('-')[0])
            self.log.debug('Mounting volume \'{0}\' from local file system to \'{1}\' in the container'.format(output, od))

        cont.image = '{0}_{1}'.format(USER, image)
        self.log.info('Adding cab \'{0}\' to recipe. The container will be named \'{1}\''.format(cont.image, name))
        self.containers.append(cont)


    def log2recipe(self, cont, recipe, num, status):

        step = {
                "name"          :   cont.name,
                "number"        :   num,
                "cab"           :   cont.image,
                "volumes"       :   cont.volumes,
                "environs"      :   cont.environs,
                "shared_memory" :   cont.shared_memory,
                "input_content" :   cont.input_content,
                "msdir_content" :   cont.msdir_content,
                "label"         :   cont.label,
                "status"        :   status,
        }
        recipe['steps'].append(step)


    def run(self, steps=None, resume=False, redo=None):
        """
        Run a Stimela recipe. 

        steps   :   recipe steps to run
        resume  :   resume recipe from last run
        redo    :   Re-run an old recipe from a .last file
        """

        recipe = {
            "name"      :   self.name,
            "steps"     :   []
        }
        start_at = 0

        if redo:
            recipe = utils.readJson(redo)
            self.log.info('Rerunning recipe {0} from {1}'.format(recipe['name'], redo))
            self.log.info('Recreating recipe instance..')
            self.containers = []
            for step in recipe['steps']:
                
            #        add I/O folders to the json file
            #        add a string describing the contents of these folders
            #        The user has to ensure that these folders exist, and have the required content
                
                self.log.info('Adding cab \'{0}\' to recipe. The container will be named \'{1}\''.format(step['cab'], step['name']))
                cont = docker.Container(step['cab'], step['name'],
                                      label=step['label'], logger=self.log,
                                      shared_memory=step['shared_memory'])
                self.log.debug('Adding volumes {0} and environmental variables {1}'.format(step['volumes'], step['environs']))
                cont.volumes = step['volumes']
                cont.environs = step['environs']
                cont.shared_memory = step['shared_memory']
                cont.input_content = step['input_content']
                cont.msdir_content = step['msdir_content']

                self.containers.append(cont)

        elif resume:
            self.log.info("Resuming recipe from last run.")
            try:
                recipe = utils.readJson(self.resume_file)
            except IOError:
                raise IOError("Cannot resume pipeline, resume file '{}' not found".format(self.resume_file))

            steps_ = recipe.pop('steps')
            recipe['steps'] = []
            _steps = []
            for step in steps_:
                if step['status'] == 'completed':
                    recipe['steps'].append(step)
                    continue

                _cab = step['cab']
                number = step['number']

                if _cab == self.containers[number-1].image:
                    self.log.info('recipe step \'{0}\' is fit for re-execution. CAB = {1}'.format(number, _cab))
                    _steps.append(number)
                else:
                    raise RuntimeError('Recipe flow, or task scheduling has changed. Cannot resume recipe. CAB= {0}'.format(_cab))
            if len(_steps)==0:
                self.log.info('All the steps were completed. No steps to resume')
                sys.exit(0)
            steps = _steps

        if getattr(steps, '__iter__', False):
            _steps = []
            if isinstance(steps[0], str):
                labels = [ cont.label.split('::')[0] for cont in self.containers]

                for step in steps:
                    try:
                        _steps.append(labels.index(step)+1)
                    except ValueError:
                        raise ValueError('Recipe label ID [{0}] doesn\'t exist'.format(step))
                steps = _steps
        else:
            steps = range(1, len(self.containers)+1)


        containers = [(step, self.containers[step-1]) for step in steps]        
        
        for i, (step, container) in enumerate(containers):
            created = False
            try:
                # Update container parameter file if need be
                if hasattr(container, '_cab'):
                    container._cab.update(container.config, container.parameter_file_name)

                self.log.info('Running Container {}'.format(container.name))
                self.log.info('STEP {0} :: {1}'.format(i+1, container.label))
                self.active = container

                container.create()
                created = True
                container.start()
                self.log2recipe(container, recipe, step, 'completed')
            except BaseException as e:
                self.completed = [cont[1] for cont in containers[:i]]
                self.remaining = [cont[1] for cont in containers[i+1:]]
                self.failed = container

                self.log.info('Recipe execution failed while running container {}'.format(container.name))
                self.log.info('Completed containers : {}'.format([c.name for c in self.completed]))
                self.log.info('Remaining containers : {}'.format([c.name for c in self.remaining]))

                self.log2recipe(container, recipe, step, 'failed')
                for step, cont in containers[i+1:]:
                    self.log.info('Logging remaining task: {}'.format(cont.label or cont.name))
                    self.log2recipe(cont, recipe, step, 'remaining')

                self.log.info('Saving pipeline information in {}'.format(self.resume_file))
                utils.writeJson(self.resume_file, recipe)

                pe = PipelineException(e, self.completed, container, self.remaining)

                raise pe, None, sys.exc_info()[2]
            finally:
                container.get_log()
                if created:
                    container.stop()
                    container.remove()

        self.proc_logger.remove('processes', self.pid)
        self.proc_logger.write()

        self.log.info('Saving pipeline information in {}'.format(self.resume_file))
        utils.writeJson(self.resume_file, recipe)

        self.log.info('Recipe executed successfully')



import Pyxis
import ms
import im
import std
from Pyxis.ModSupport import *
import im.lwimager
import im.argo
import pyfits

import tempfile
import json
import numpy
import os

import pyrap.measures
dm = pyrap.measures.measures()

def readJson(conf):

    with open(conf) as _std:
        jdict = json.load(_std)

    for key,val in jdict.iteritems():
        if isinstance(val, unicode):
            jdict[key] = str(val)

    return jdict

OUTDIR = os.environ["OUTPUT"]
INDIR = os.environ["INPUT"]
CONFIG = os.environ["CONFIG"]
MSDIR = os.environ["MSDIR"]
LOG_Template = "${OUTDIR>/}log-lwimager_predict.txt"
v.OUTFILE = II("${OUTDIR>/}results-${MS:BASE}")


def azishe():

    # Get parameters
    jdict = readJson(CONFIG)

    v.MS = "{:s}/{:s}".format(MSDIR, jdict.get("msname"))
    v.LOG_Template = "${OUTDIR>/}log-${MS:BASE}-lwimager_predict.txt"

    for item in [INDIR, "/data/skymodels/"]:
        lsmname = "{:s}/{:s}".format(item, jdict["skymodel"])
        if os.path.exists(lsmname):
            break

    v.LSM = II("${lsmname:FILE}")
    
    std.copy(lsmname, LSM)

    adaptFITS(LSM)

    if jdict["recentre"]:
        direction = jdict["direction"].split(",")
        radec = dm.direction(*direction)

        with pyfits.open(LSM) as hdu:
            hdu[0].header["crval1"] = numpy.rad2deg( radec["m0"]["value"] )
            hdu[0].header["crval2"] = numpy.rad2deg( radec["m1"]["value"] )
            hdu.writeto(LSM, clobber=True)

    column = jdict.get("column", "DATA")
    if column not in ["DATA", "CORRECTED_DATA", "MODEL_DATA"]:
        im.argo.addcol(colname=column)

    add_to_col = jdict.get("add_to_column", None)
    copy = jdict.get("copy_to_CORRECTED_DATA", False)
    addnoise = jdict.get("addnoise", True)

    ms.set_default_spectral_info()


    if add_to_col:
        im.lwimager.predict_vis(image=LSM, padding=1.5, wprojplanes=128, column="MODEL_DATA")

        tab = ms.msw()
        data = tab.getcol("MODEL_DATA") + tab.getcol(add_to_col)
        tab.putcol(column, data)

        tab.close() 
    else:
        im.lwimager.predict_vis(image=LSM, padding=1.5, wprojplanes=128, column=column)


    if addnoise:
        sefd = jdict.get("sefd", 551) # default is MeerKAT L-Band
        noise = compute_vis_noise(sefd)
        simnoise(noise, addToCol=column, column=column)

    if copy and column!="CORRECTED_DATA":
        ms.copycol(fromcol=column, tocol="CORRECTED_DATA")


def simnoise (noise=0, rowchunk=100000, 
              addToCol=None, scale_noise=1.0, 
              column='MODEL_DATA'):
    """ Simulate noise into an MS """

    spwtab = ms.ms(subtable="SPECTRAL_WINDOW")
    freq0 = spwtab.getcol("CHAN_FREQ")[ms.SPWID,0]/1e6

    tab = ms.msw()
    dshape = list(tab.getcol('DATA').shape)
    nrows = dshape[0]

    noise = noise or compute_vis_noise()

    if addToCol: colData = tab.getcol(addToCol)

    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        dshape[0] = nr
        data = noise*(numpy.random.randn(*dshape) + 1j*numpy.random.randn(*dshape)) * scale_noise

        if addToCol: 
            data+=colData[row0:(row0+nr)]
            info(" $addToCol + noise --> $column (rows $row0 to %d)"%(row0+nr-1))
        else : info("Adding noise to $column (rows $row0 to %d)"%(row0+nr-1))

        tab.putcol(column, data, row0, nr)
    tab.close() 



def compute_vis_noise(sefd):
    """Computes nominal per-visibility noise"""

    tab = ms.ms()
    spwtab = ms.ms(subtable="SPECTRAL_WINDOW")

    freq0 = spwtab.getcol("CHAN_FREQ")[ms.SPWID, 0]
    wavelength = 300e+6/freq0
    bw = spwtab.getcol("CHAN_WIDTH")[ms.SPWID, 0]
    dt = tab.getcol("EXPOSURE", 0, 1)[0]
    dtf = (tab.getcol("TIME", tab.nrows()-1, 1)-tab.getcol("TIME", 0, 1))[0]

    # close tables properly, else the calls below will hang waiting for a lock...
    tab.close()
    spwtab.close()

    info(">>> $MS freq %.2f MHz (lambda=%.2fm), bandwidth %.2g kHz, %.2fs integrations, %.2fh synthesis"%(freq0*1e-6, wavelength, bw*1e-3, dt, dtf/3600))
    noise = sefd/math.sqrt(abs(2*bw*dt))
    info(">>> SEFD of %.2f Jy gives per-visibility noise of %.2f mJy"%(sefd, noise*1000))

    return noise 


def adaptFITS(imagename):
    """ Try to re-structre FITS file so that it conforms to lwimager standard """
    
    hdr = pyfits.open(imagename)[0].header
    naxis = hdr["NAXIS"]
    
    # Figure out if any axes have to be added before we proceed
    if naxis>=2 and naxis < 4:
        _freq = "--add-axis=freq:$FREQ0:DFREQ:Hz"
        _stokes = "--add-axis=stokes:1:1:1:Jy/Beam"
    
        if naxis == 2:
            info("FITS Image has 2 axes. Will add FREQ and STOKES axes. We need these predict visibilities")
            x.sh("fiitstool.py ${_stokes} ${_freq} $imagename")

        elif naxis==3:
            if hdr["CTYPE3"].lower().startswith("freq"):
                info("FITS Image missing FREQ axis. Adding it")
                # Will also need to reorder if freq is 3rd axis
                x.sh("fitstool.py ${_stokes} $imagename && fitstool.py --reorder=1,2,4,3 $imagename")

            elif hdr["CTYPE3"].lower().startswith("stokes"):
                info("FITS Image missing STOKES axis. Adding it")
                x.sh("fitstool.py ${_freq} $imagename")

    with pyfits.open(imagename) as hdu:
        hdr = hdu[0].header
        freq_ind = filter(lambda ind: hdr["CTYPE%d"%ind].startswith("FREQ"), range(1,5) )[0]

        if hdr["CDELT%d"%freq_ind] <0:
            dfreq = abs(hdr["CDELT%d"%freq_ind])
            nchan = hdr["NAXIS%d"%freq_ind]
            freq0 = hdr["CRVAL%d"%freq_ind]
            hdu[0].header["CDELT%d"%freq_ind] = dfreq
            hdu[0].header["CRVAL%d"%freq_ind] = freq0 - nchan*dfreq
            hdu.writeto(imagename, clobber=True)
        
    info("You image is now RATT approved ;) ")



# This is my pyxis tool box

import Pyxis
import pyfits
import numpy
import math
from pyrap.tables import table


def compute_vis_noise(sefd):
    """Computes nominal per-visibility noise"""

    tab = ms.ms()
    spwtab = ms.ms(subtable="SPECTRAL_WINDOW")

    freq0 = spwtab.getcol("CHAN_FREQ")[ms.SPWID, 0]
    wavelength = 300e+6/freq0
    bw = spwtab.getcol("CHAN_WIDTH")[ms.SPWID, 0]
    dt = tab.getcol("EXPOSURE", 0, 1)[0]
    dtf = (tab.getcol("TIME", tab.nrows()-1, 1)-tab.getcol("TIME", 0, 1))[0]

    # close tables properly, else the calls below will hang waiting for a lock...
    tab.close()
    spwtab.close()

    info(">>> $MS freq %.2f MHz (lambda=%.2fm), bandwidth %.2g kHz, %.2fs integrations, %.2fh synthesis"%(freq0*1e-6, wavelength, bw*1e-3, dt, dtf/3600))
    noise = sefd/math.sqrt(abs(2*bw*dt))
    info(">>> SEFD of %.2f Jy gives per-visibility noise of %.2f mJy"%(sefd, noise*1000))

    return noise 


def simnoise (noise=0, rowchunk=100000, 
              addToCol=None, scale_noise=1.0, 
              column='MODEL_DATA'):
    """ Simulate noise into an MS """

    spwtab = ms.ms(subtable="SPECTRAL_WINDOW")
    freq0 = spwtab.getcol("CHAN_FREQ")[ms.SPWID,0]/1e6

    tab = ms.msw()
    dshape = list(tab.getcol('DATA').shape)
    nrows = dshape[0]

    noise = noise or compute_vis_noise()

    if addToCol: colData = tab.getcol(addToCol)

    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        dshape[0] = nr
        data = noise*(numpy.random.randn(*dshape) + 1j*numpy.random.randn(*dshape)) * scale_noise

        if addToCol: 
            data+=colData[row0:(row0+nr)]
            info(" $addToCol + noise --> $column (rows $row0 to %d)"%(row0+nr-1))
        else : info("Adding noise to $column (rows $row0 to %d)"%(row0+nr-1))

        tab.putcol(column, data, row0, nr)
    tab.close() 


def adaptFITS(imagename):
    """ Try to re-structre FITS file so that it conforms to lwimager standard """
    
    hdr = pyfits.open(imagename)[0].header
    naxis = hdr["NAXIS"]
    
    # Figure out if any axes have to be added before we proceed
    if naxis>=2 and naxis < 4:
        _freq = "--add-axis=freq:$FREQ0:DFREQ:Hz"
        _stokes = "--add-axis=stokes:1:1:1:Jy/Beam"
    
        if naxis == 2:
            info("FITS Image has 2 axes. Will add FREQ and STOKES axes. We need these predict visibilities")
            x.sh("fiitstool.py ${_stokes} ${_freq} $imagename")

        elif naxis==3:
            if hdr["CTYPE3"].lower().startswith("freq"):
                info("FITS Image missing FREQ axis. Adding it")
                # Will also need to reorder if freq is 3rd axis
                x.sh("fitstool.py ${_stokes} $imagename && fitstool.py --reorder=1,2,4,3 $imagename")

            elif hdr["CTYPE3"].lower().startswith("stokes"):
                info("FITS Image missing STOKES axis. Adding it")
                x.sh("fitstool.py ${_freq} $imagename")

    with pyfits.open(imagename) as hdu:
        hdr = hdu[0].header
        freq_ind = filter(lambda ind: hdr["CTYPE%d"%ind].startswith("FREQ"), range(1,5) )[0]

        if hdr["CDELT%d"%freq_ind] <0:
            dfreq = abs(hdr["CDELT%d"%freq_ind])
            nchan = hdr["NAXIS%d"%freq_ind]
            freq0 = hdr["CRVAL%d"%freq_ind]
            hdu[0].header["CDELT%d"%freq_ind] = dfreq
            hdu[0].header["CRVAL%d"%freq_ind] = freq0 - nchan*dfreq
            hdu.writeto(imagename, clobber=True)
        
    info("You image is now RATT approved ;) ")



import inner_taper as taper
import os
import sys

sys.path.append("/utils")
import utils

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

jdict = utils.readJson(CONFIG)

msname = "%s/%s"%(MSDIR, jdict["msname"])
res = jdict["res"]
freq = jdict["freq"]
reset = jdict.get("reset", False)


savefig = "%s/%s_%.2f-%.2f.png"%(OUTPUT, jdict["msname"][:-3], res[0], res[1])

taper.taper(msname, res=res, freq=freq, savefig=savefig)

if reset:
    taper.reset(msname)



import os
import sys

sys.path.append('/utils')
import utils


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
params = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value in [False, None]:
        continue

    if value is True:
        value = ""

    # Positional arguments
    if name == 'input-skymodel':
        inlsm= value
        continue
    elif name == 'tag':
        tag = value
        continue

    params[name] = value

#TODO: Need fix tigger-tag, these kludges are annoying
if params.pop('transfer-tags', False) in [True, ""]:
    if params.get('tolerance', None) is None:
        raise RuntimeError("Parameter 'tolerance' is required when 'transfer-tags' is enables")
    args = ['{0}transfer-tags {1}:{2}'.format(cab['prefix'], inlsm, params.pop('tolerance'))]
    inlsm = params.get('output')
else:
    args = []

args += ['{0}{1} {2}'.format(cab['prefix'], name, value) for name, value in params.iteritems()]
utils.xrun(cab['binary'], args+[inlsm, tag])



import os
import sys
import json
import pyfits
import codecs
import subprocess
from lofar import bdsm
import Tigger
import copy
import numpy
from astLib.astWCS import WCS

from Tigger.Models import SkyModel, ModelClasses

sys.path.append("/utils")
import utils

CONFIG = os.environ["CONFIG"]
INDIR = os.environ["INPUT"]
OUTDIR = os.environ["OUTPUT"]

jdict = utils.readJson(CONFIG)
template = utils.readJson("/code/sourcery_template.json")


name = jdict.pop("imagename")
dimage = jdict.pop("detection_image", None)
psf = jdict.pop("psf", None)
prefix = jdict.pop("prefix", None) or name.split("/")[-1]

for item in "name psf dimage prefix".split():
    if isinstance(globals()[item], str):
        tmp = OUTDIR + "/" + globals()[item]  if item=="prefix" else INDIR + "/" + globals()[item]
        globals()[item] = utils.substitute_globals(globals()[item]) or tmp


def pybdsm(image, prefix, *args, **kw):
    img = bdsm.process_image(image, *args, **kw)
    gaulfile = prefix+".gaul.fits"
    img.write_catalog(catalog_type="gaul", format="fits", outfile=gaulfile, clobber=True)

    # convert to Gaul file to Tigger LSM
    # First make dummy tigger model
    tname = prefix.split("/")[-1] + ".txt"
    tname_lsm = prefix + ".lsm.html"
    with open(tname, "w") as stdw:
        stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

    model = Tigger.load(tname)

    def tigger_src(gaul, idx):

        name = "SRC%d"%idx

        flux = ModelClasses.Polarization(gaul["Total_flux"], 0, 0, 0, I_err=gaul["E_Total_flux"])

        ra, ra_err = map(numpy.deg2rad, (gaul["RA"], gaul["E_RA"]) )
        dec, dec_err = map(numpy.deg2rad, (gaul["DEC"], gaul["E_DEC"]) )
        pos =  ModelClasses.Position(ra, dec, ra_err=ra_err, dec_err=dec_err)

        ex, ex_err = map(numpy.deg2rad, (gaul["DC_Maj"], gaul["E_DC_Maj"]) )
        ey, ey_err = map(numpy.deg2rad, (gaul["DC_Min"], gaul["E_DC_Min"]) )
        pa, pa_err = map(numpy.deg2rad, (gaul["PA"], gaul["E_PA"]) )

        if ex and ey:
            shape = ModelClasses.Gaussian(ex, ey, pa, ex_err=ex_err, ey_err=ey_err, pa_err=pa_err)
        else:
            shape = None
      
        return SkyModel.Source(name, pos, flux, shape=shape)

        
    with pyfits.open(gaulfile) as hdu:
        data = hdu[1].data

        for i, gaul in enumerate(data):
            model.sources.append(tigger_src(gaul, i))

    wcs = WCS(image)
    centre = wcs.getCentreWCSCoords()
    model.ra0, model.dec0 = map(numpy.deg2rad, centre)

    model.save(tname_lsm)
    # Rename using CORPAT
    utils.xrun("tigger-convert", [tname_lsm, "--rename -f"])




go = jdict.pop("pybdsm", False)

if go:
    if dimage:
        jdict["detection_image"] = dimage
    pybdsm(name, prefix, **jdict)
    sys.exit(0)

template["prefix"] = prefix
template["imagename"] = name

template["psfname"] = psf
template["reliability"]["thresh_pix"] = jdict.pop("thresh_pix", 4)
template["reliability"]["thresh_isl"] = jdict.pop("thresh_isl", 3)
if dimage:
    template["reliability"]["detection_image"] = dimage


# The tagging requires too much fine tunning. Turning it off by default
template["dd_tagging"]["enable"] = jdict.pop("dd_tagging", False)



for key in jdict:
    if key in template["reliability"].keys():
        template["reliability"][key] = jdict[key]
    elif key in template["dd_tagging"].keys():
        template["dd_tagging"][key] = jdict[key]


config = "run_me_now.json"
utils.writeJson(config, template)

utils.xrun("sourcery", ["-jc", config])
utils.xrun("tigger-convert", ["%s/%s.lsm.html"%(OUTDIR,prefix), "--rename -f"])

utils.xrun("rm", ["-f", config])



import os
import sys
import numpy
from pyrap.tables import table

sys.path.append("/utils")
import utils

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

jdict = dict(npix=2048, nfacets=11, MaxMajorIter=5, cellsize=5, Enable=False, weight="briggs", robust=0,
             WeightCol="WEIGHT_SPECTRUM")

jdict.update( utils.readJson(CONFIG) )

def find_closest(A, target):
    #A must be sorted
    idx = A.searchsorted(target)
    idx = numpy.clip(idx, 1, len(A)-1)
    left = A[idx-1]
    right = A[idx]
    idx -= target - left < right - target
    return idx


STANDARD_OPTS = {
    "field_id" : "Field",
    "spw_id" : "DDID",
    "imageprefix" : "ImageName",
    "prefix"    :   "ImageName",
    "weight" : "Weighting",
    "robust" : "Robust",
    "npix" : "Npix",
    "cellsize": "Cell",
    "nfacets" : "NFacets",
    "column" : "ColName",
    "clean_mask" : "CleanMaskImage",
}

niter = jdict.pop("clean_iterations", 0)
niter = jdict.pop("MaxMinorIter", niter)
parset = jdict.pop("parset", None)

if parset:
    parset = ["%s/%s"%(INPUT, parset)]
else:
    parset = []

mode = "Dirty"
if niter:
    mode = "Clean"

addToCol = str(jdict.pop("add_to_column", None))

if jdict.pop("predict", False):
    skymodel = jdict.pop("skymodel", None)
    if skymodel:
        mode = "Predict"
        skymodel = INPUT + "/" + skymodel
        column = str(jdict.get("column", None))
        if column == None:
            column = str(jdict.get("ColName", None))
        if column == None:
            jdict["column"] = "MODEL_DATA"

beam_pattern = jdict.pop("beam_files_pattern", None)
msname = jdict.pop("msname")
if not isinstance(msname, (tuple, list)):
    msname = [msname]
prefix = jdict.pop("imageprefix", os.path.basename(msname[0])[:-3])

options = {}
for key, value in jdict.iteritems():
    key = STANDARD_OPTS.get(key, key)

    if value:
        options[key] = value
    if key in options and options[key] == None:
        del options[key]

mask = options.get("CleanMaskImage", None)

for item in "mask prefix beam_pattern".split():
    if isinstance(globals()[item], str):
        globals()[item] = utils.substitute_globals(globals()[item]) or "%s/%s"(OUTPUT if item=="prefix" else INPUT, globals()[item])

with open("mslist.txt", "w") as std:
    for ms in msname:
        std.write("{:s}/{:s}\n".format(MSDIR, ms))

options["MSName"] = "mslist.txt"
options["ImageName"] = prefix
options["SaveImages"] = "all"
if niter:
    options["MaxMinorIter"] = niter

column = options.get("ColName", "MODEL_DATA")

upfirst = lambda a: a[0].upper() + a[1:]
options["Weighting"] = upfirst(options["Weighting"])

if options.pop("add_beam", False):
    if beam_pattern:
        options["BeamModel"] = "FITS"
        options["FITSLAxis"] = options.pop("beam_l_axis", "L")
        options["FITSMAxis"] = options.pop("beam_m_axis", "M")
        options["FITSFile"] = "'%s'"%beam_pattern


if mode.lower()=="predict":
    import pyfits
    
    odds = numpy.array([65,75,77,81,91,99,105,117,125,135,143,147,165,175,189,195,225,231,243,245,273,275,297,315,325,343,351,375,385,405,429,441,455,495,525,539,567,585,625,637,675,693,715,729,735,819,825,875,891,945,975,1001,1029,1053,1125,1155,1215,1225,1287,1323,1365,1375,1485,1575,1617,1625,1701,1715,1755,1875,1911,1925,2025,2079,2145,2187,2205,2275,2401,2457,2475,2625,2673,2695,2835,2925,3003,3087,3125,3159,3185,3375,3465,3575,3645,3675,3773,3861,3969,4095,4125,4375,4455,4459,4725,4851,4875,5005,5103,5145,5265,5625,5733,5775,6075,6125,6237,6435,6561,6615,6825,6875,7007,7203,7371,7425,7875,8019,8085,8125,8505,8575,8775,9009,9261,9375,9477,9555,9625,10125,10395,10725,10935,11025,11319,11375,11583,11907,12005,12285,12375,13125,13365,13377,13475,14175,14553,14625,15015,15309,15435,15625,15795,15925,16807,16875,17199,17325,17875,18225,18375,18711,18865,19305,19683,19845])
    
    hdu = pyfits.open(skymodel)
    nx = hdu[0].header["naxis1"]
    ny = hdu[0].header["naxis2"]
    ndim = hdu[0].header["naxis"]
    cell = abs(hdu[0].header["cdelt1"])*3600
    ra0 = hdu[0].header["crval1"]
    dec0 = hdu[0].header["crval2"]

    recenter = options.pop("recenter", False)
        
    idx = find_closest(odds, nx)
    if odds[idx]<nx:
        tx = odds[idx + 1]
    else:
        tx = odds[idx]

    idy = find_closest(odds, ny)
    if odds[idy]<ny:
        ty = odds[idy + 1]
    else:
        ty = odds[idy]

    pad_width = numpy.zeros([ndim,2], dtype=int).tolist()

    dfx = tx - nx
    if dfx%2!=0:
        ax = dfx - dfx/2 * 2
    else:
        ax = 0
    pad_width[-1] = (dfx/2, dfx/2 + ax)


    dfy = ty - ny
    if dfx%2!=0:
        ay = dfy - dfy/2 * 2
    else:
        ay = 0
    pad_width[-2] = (dfy/2, dfy/2 + ay)

    data = numpy.pad( hdu[0].data, 
                             pad_width=pad_width, 
                             mode="constant", constant_values=0)

    hdu.close()

    options["Mode"] = "Dirty"
    options["Npix"] = data.shape[-1]
    options["Cell"] = cell

    cmd = [ "--%s=%s"%(key, value) for (key, value) in options.iteritems() ]
    utils.xrun("/ddfvenv/bin/DDF.py", parset + cmd)

    with pyfits.open("%s.dirty.fits"%prefix) as hdu:
        hdu[0].data = data
        if recenter:
            pass
        else:
            hdu[0].header["crval1"] = ra0
            hdu[0].header["crval2"] = dec0

        hdu[0].writeto("%s.dirty.fits"%prefix, clobber=True)

    options["Mode"] = "Predict"
    options["PredictModelName"] = "%s.dirty.fits"%prefix

    cmd = [ "--%s=%s"%(key, value) for (key, value) in options.iteritems() ]
    utils.xrun("/ddfvenv/bin/DDF.py", parset + cmd)

    if column != "MODEL_DATA":
        print("Copying data from MODEL_DATA to %s"%(column))
        for ms_ in map(str, msname):
            msn = "%s/%s"%(MSDIR, ms_)
            tab = table(msn, readonly=False)
            column = str(column)
            data = tab.getcol("MODEL_DATA")

            if addToCol:
                data = tab.getcol(addToCol) + data

            nrows = tab.nrows()
            rowchunk = nrows//10
            
            for row0 in range(0, nrows, rowchunk):
                nr = min(rowchunk, nrows-row0)
                print("Moving data form MODEL_DATA to %s (rows %d to %d)"%(column, row0, row0+nr-1))
                tab.putcol(column, data[row0:(row0+nr)], row0, nr)

            tab.close()

else:
    options["Mode"] = mode
    cmd = [ "--%s=%s"%(key, value) for (key, value) in options.iteritems() ]
    utils.xrun("/ddfvenv/bin/DDF.py", parset + cmd)



from __future__ import print_function
import os
import sys

import drivecasa
casa = drivecasa.Casapy(log2term=True, echo_to_stdout=True, timeout=24*3600)

sys.path.append('/utils')
import utils

CONFIG = os.environ['CONFIG']
cab = utils.readJson(CONFIG)

params = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    params[name] = value

port2fits = params.pop('port2fits', True)
keep_casa_images = params.pop("keep_casa_images", False)
script = ['clean(**{})'.format(params)]

nterms = params.get("nterms", 1)
images = ["flux", "model", "residual", "psf", "image"]
STD_IMAGES = images[:4]

def log2term(result):
    if result[0]:
        out = '\n'.join(result[0])
        sys.stdout.write(out)
    if result[1]:
        err = '\n'.join(result[1] if result[1] else [''])
        failed = err.lower().find('an error occurred running task')>=0
        if failed:
            raise RuntimeError('CASA Task failed. See error message bellow \n {}'.format(err))
        sys.stdout.write('WARNING:: SEVERE messages from CASA run: \n {}'.format(err))

result = casa.run_script(script, raise_on_severe=False)
log2term(result)

prefix = params['imagename']
convert = []
if port2fits:
    for image in images:
        img ="{:s}.{:s}".format(prefix, image)
        if image == 'flux':
            _images = [img]
        elif nterms>1:
            _images = ["%s.tt%d"%(img,d) for d in range(nterms)]
            if image=="image":
                if  nterms==2:
                    alpha = img+".alpha"
                    alpha_err = img+".alpha.error"
                    _images += [alpha,alpha_err]
                if  nterms==3:
                    beta = img+".beta"
                    beta_err = img+".beta.error"
                    _images += [beta,beta_err]
        else:
            _images = [img]
        convert += _images

script = []
for _image in  convert:
    sys.stdout.write(_image)
    if _image in STD_IMAGES and (not os.path.exists(_image)):
        raise RuntimeError("Standard output from CASA clean task not found. Something went wrong durring cleaning, take a look at the logs and such")

    elif os.path.exists(_image):
        script += ['exportfits(**{})'.format( dict(imagename=_image, fitsimage=_image+".fits", overwrite=True) )]

result = casa.run_script(script, raise_on_severe=False)
log2term(result)

if not keep_casa_images:
    for _image in convert:
        utils.xrun("rm", ["-rf", _image])



import os
import sys
import json
import pyfits
import codecs

CONFIG = os.environ["CONFIG"]
INDIR = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

outfile = sys.argv[1]

_ANTENNAS = {
     "meerkat": "meerkat.itrf.txt",
     "kat-7": "kat-7.itrf.txt",
     "jvlaa": "vlaa.itrf.txt",
     "jvlab": "vlab.itrf.txt",
     "jvlac": "vlac.itrf.txt",
     "jvlad": "vlad.itrf.txt",
     "wsrt": "wsrt.itrf.txt",
     "ska1mid254": "skamid254.itrf.txt",
     "ska1mid197": "skamid197.itrf.txt",
}

_OBS = {
     "meerkat": "meerkat",
     "kat-7": "kat-7",
     "jvlaa": "vla",
     "jvlab": "vla",
     "jvlac": "vla",
     "jvlad": "vla",
     "wsrt": "wsrt",
     "ska1mid254": "meerkat",
     "ska1mid197": "meerkat",
}



def readJson(conf):

    with open(conf) as _std:
        jdict = json.load(_std)

    out = {}

    for key,val in jdict.iteritems():

        if isinstance(val, unicode):
            val = str(val)

        out[str(key)] = val

    return out


direction = jdict.get("direction", None)
jdict = readJson(CONFIG)
telescope = jdict["telescope"]

jdict["outdir"] = MSDIR
jdict["tel"] = _OBS[telescope]

if jdict.get("antennas", False):
    jdict["pos"] = INDIR+"/"+jdict.get("antennas", "meetkat")
    jdict["coords"] = jdict.get("coord_sys", "enu")
else:
    jdict["pos"] = "/data/observatories/"+_ANTENNAS[telescope]
    if not os.path.isdir(jdict["pos"]):
        jdict["pos_type"] = "ascii"
        jdict["coords"] = "itrf"

for item in ["antennas", "coord_sys", "telescope"]:
    jdict.pop(item, None)

imagename = jdict.pop("skymodel", False)

if jdict.pop("predict", False) and imagename:

    for item in [INDIR, "/data/skymodels"]:
        _imagename = "{:s}/{:s}".format(item, imagename)
        if os.path.exists(_imagename):
            break

    imagename = _imagename

    with pyfits.open(imagename) as hdu:
        hdr = hdu[0].header

        freq = filter(lambda ind: hdr["ctype%d"%ind].startswith("FREQ"),
                range(2, 5, 1))
        if freq:
            freq = freq[0]
        else:
            raise TypeError("Your FITS image has no frequency information")

        nchan = hdr["naxis%d"%freq]
        freq0 = hdr["crval%d"%freq]
        dfreq = hdr["cdelt%d"%freq]

        jdict["freq0"] = freq0
        jdict["dfreq"] = dfreq
        jdict["nchan"] = nchan

        # get direction
        if not direction:
            ra0 = hdr["crval1"]
            dec0 = hdr["crval2"]

            jdict["direction"] = "J2000,%fdeg,%fdeg"%(ra0, dec0)
            raise SystemExit(direction)

with codecs.open(outfile, "w", "utf8") as std:
    std.write( json.dumps(jdict, ensure_ascii=False) )

raise SystemExit("adfhgsdjsadjfkhsd sdfkhgsdgfhsgf")



import os
import sys
import re
from lofar import bdsm
import numpy
import Tigger
import tempfile
import pyfits

from astLib.astWCS import WCS
from Tigger.Models import SkyModel, ModelClasses


sys.path.append('/utils')
import utils

CONFIG = os.environ['CONFIG']
INPUT = os.environ['INPUT']
OUTPUT = os.environ['OUTPUT']
MSDIR = os.environ['MSDIR']

cab = utils.readJson(CONFIG)

write_catalog = ['bbs_patches', 'bbs_patches_mask',
            'catalog_type', 'clobber', 'correct_proj', 'format', 
            'incl_chan', 'incl_empty', 'srcroot', 'port2tigger', 'outfile']

img_opts = {}
write_opts = {}

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    if name in write_catalog:
        write_opts[name] = value
    else:
        img_opts[name] = value

image = img_opts.pop('filename')
outfile = write_opts.pop('outfile')

img = bdsm.process_image(image, **img_opts)

port2tigger = write_opts.pop('port2tigger', True)

if write_opts.get('format', None) !='fits':
    write_opts['format'] = 'fits'

img.write_catalog(outfile=outfile, **write_opts)

if not port2tigger:
    sys.exit(0)

# convert to Gaul file to Tigger LSM
# First make dummy tigger model
tfile = tempfile.NamedTemporaryFile(suffix='.txt')
tfile.flush()

prefix = outfile[:-5]
tname_lsm = prefix + ".lsm.html"
with open(tfile.name, "w") as stdw:
    stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

model = Tigger.load(tfile.name)
tfile.close()

def tigger_src(src, idx):

    name = "SRC%d"%idx

    flux = ModelClasses.Polarization(src["Total_flux"], 0, 0, 0, I_err=src["E_Total_flux"])

    ra, ra_err = map(numpy.deg2rad, (src["RA"], src["E_RA"]) )
    dec, dec_err = map(numpy.deg2rad, (src["DEC"], src["E_DEC"]) )
    pos =  ModelClasses.Position(ra, dec, ra_err=ra_err, dec_err=dec_err)

    ex, ex_err = map(numpy.deg2rad, (src["DC_Maj"], src["E_DC_Maj"]) )
    ey, ey_err = map(numpy.deg2rad, (src["DC_Min"], src["E_DC_Min"]) )
    pa, pa_err = map(numpy.deg2rad, (src["PA"], src["E_PA"]) )

    if ex and ey: 
        shape = ModelClasses.Gaussian(ex, ey, pa, ex_err=ex_err, ey_err=ey_err, pa_err=pa_err)
    else:
        shape = None

    return SkyModel.Source(name, pos, flux, shape=shape)


with pyfits.open(outfile) as hdu:
    data = hdu[1].data

    for i, src in enumerate(data):
        model.sources.append(tigger_src(src, i)) 

wcs = WCS(image)
centre = wcs.getCentreWCSCoords()
model.ra0, model.dec0 = map(numpy.deg2rad, centre)

model.save(tname_lsm)
# Rename using CORPAT
utils.xrun("tigger-convert", [tname_lsm, "--rename -f"])



import os
import sys
import argparse
import time
from  argparse import ArgumentParser
import textwrap as _textwrap
import tempfile
import signal

import inspect
import stimela
from stimela_misc import version
from stimela import utils, cargo
from stimela.cargo import cab
from recipe import Recipe as Pipeline
from recipe import Recipe, PipelineException
from stimela import docker

from stimela.utils import logger

LOG_HOME = os.path.expanduser("~/.stimela")
LOG_FILE = LOG_HOME + "/stimela_logfile.json"


BASE = os.listdir(cargo.BASE_PATH)
CAB = []
for item in os.listdir(cargo.CAB_PATH):
    try:
        dockerfile = 'Dockerfile' in os.listdir('{0}/{1}'.format(cargo.CAB_PATH, item))
    except OSError:
        continue
    if dockerfile:
        CAB.append(item)


NOT_PUBLIC = ["ddfacet"]

USER = os.environ["USER"]
UID = os.getuid()
GID = os.getgid()

container_home = '/home/{}'.format(USER)

__version__ = version.version

GLOBALS = {}

class MultilineFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n ')
        multiline_text = ''
        for paragraph in paragraphs:
            formatted_paragraph = _textwrap.fill(paragraph, width, initial_indent=indent, subsequent_indent=indent) + '\n\n'
            multiline_text = multiline_text + formatted_paragraph
        return multiline_text

def register_globals():
    frame = inspect.currentframe().f_back
    frame.f_globals.update(GLOBALS)

def build(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv = ' ' + arg

    parser = ArgumentParser(description='Build executor (a.k.a cab) images')
    parser.add_argument("-b", "--base", action="store_true",
            help="Build base images")

    parser.add_argument("-c", "--cab", metavar="CAB,CAB_DIR",
            help="Executor image (name) name, location of executor image files")

    parser.add_argument("-uo", "--us-only",
            help="Only build these cabs. Comma separated cab names")

    parser.add_argument("-i", "--ignore-cabs", default="",
            help="Comma separated cabs (executor images) to ignore.")

    parser.add_argument("-nc", "--no-cache", action="store_true",
            help="Do not use cache when building the image")

    args = parser.parse_args(argv)
    log = logger.StimelaLogger(LOG_FILE)

    if args.base:
        for image in BASE:
            dockerfile = "{:s}/{:s}".format(cargo.BASE_PATH, image)
            image = "stimela/{0}:{1}".format(image, __version__)
            docker.build(image,
                         dockerfile)

        log.log_image(image, replace=True)
        log.write()

        return 0

    workdir = "/home/{}/output/".format(USER)
    build_args = ["RUN groupadd -g {:d} {:s}".format(UID, USER),
                  "RUN useradd -u {:d} -g {:d} {:s}".format(UID, UID, USER),
                  "WORKDIR {:s}".format(workdir),
                  "ENV HOME {:s}".format(USER),
                  "USER {:s}".format(USER)]

    no_cache = ["--no-cache"] if args.no_cache else []


    if args.cab:
        cab_args = args.cab.split(",")

        if len(cab_args)==2:
            cab, path = cab_args
        else:
            raise ValueError("Not enough arguments for build command.")

        image = "{:s}_cab/{:s}".format(USER, cab)

        docker.build(cab,
                     path,
                     build_args=build_args, args=no_cache)

        log.log_image(image, replace=True, cab=True)
        log.write()
        return

    for image in CAB:
        IGNORE = args.ignore_cabs.split(",")
        if image in NOT_PUBLIC+IGNORE:
            continue

        dockerfile = "{:s}/{:s}".format(cargo.CAB_PATH, image)
        image = "{:s}_cab/{:s}".format(USER, image)
        docker.build(image,
                     dockerfile,
                     build_args=build_args, args=no_cache)

        log.log_image(image, replace=True, cab=True)
        log.write()



def info(cabname, header=False):
    """ prints out help information about a cab """

    # First check if cab exists
    if cabname not in CAB:
        raise RuntimeError("Cab '{}' could not be found. The available cabs are listed below \n {}".format(cabname, CAB))

    pfile = "{0}/{1}/parameters.json".format(cargo.CAB_PATH, cabname)
    cab_definition = cab.CabDefinition(parameter_file=pfile)
    cab_definition.display(header)


def cabs(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv = ' ' + arg

    parser = ArgumentParser(description='List executor (a.k.a cab) images')
    parser.add_argument("-i", "--cab-doc", 
        help="Will display document about the specified cab. For example, \
to get help on the 'cleanmask cab' run 'stimela cabs --cab-doc cleanmask'")

    parser.add_argument("-ls", "--list", action="store_true",
            help="List cabs")

    args = parser.parse_args(argv)

    if args.cab_doc:
        info(args.cab_doc)
    else:
        for cab in CAB:
            try:
                info(cab, header=True)
            except IOError:
                pass


def run(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Dockerized Radio Interferometric Scripting Framework.\n'
                                        'Sphesihle Makhathini <sphemakh@gmail.com>')

    add = parser.add_argument

    add("-in", "--input",
            help="Input folder")

    add("-out", "--output", default="output",
            help="Output folder")

    add("-ms", "--msdir",
            help="MS folder. MSs should be placed here. Also, empty MSs will be placed here")

    add("-j", "--ncores", type=int,
            help="Number of cores to when stimela parallesization (stimea.utils.pper function) ")

    add("script",
            help="Run script")

    add("-g", "--globals", metavar="KEY=VALUE[:TYPE]", action="append", default=[],
            help="Global variables to pass to script. The type is assumed to string unless specified")

    args = parser.parse_args(argv)
    tag =  None

    _globals = dict(STIMELA_INPUT=args.input, STIMELA_OUTPUT=args.output,
                    STIMELA_MSDIR=args.msdir,
                    CAB_TAG=tag)

    nargs = len(args.globals)

    global GLOBALS

    if nargs:
        for arg in args.globals:
            if arg.find("=") > 1:
                key, value = arg.split("=")

                try:
                    value, _type = value.split(":")
                except ValueError:
                    _type = "str"

                GLOBALS[key] = eval("{:s}('{:s}')".format(_type, value))

    if args.ncores:
        utils.CPUS = args.ncores

    execfile(args.script, _globals)


def pull(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Pull docker stimela base images')

    add = parser.add_argument

    add("-image", action="append", metavar="IMAGE[:TAG]",
            help="Pull base image along with its tag (or version). Can be called multiple times")

    add("-t", "--tag",
            help="Tag")

    args = parser.parse_args(argv)
    log = logger.StimelaLogger(LOG_FILE)

    if args.image:
        for image in args.image:

            if not img.find(image):
                docker.pull(image)
                log.log_image(image)
    else:

        base = []
        for cab in CAB:
            image = "{:s}/{:s}".format(cargo.CAB_PATH, cab)
            base.append( utils.get_Dockerfile_base_image(image).split()[-1] )

        base = set(base)

        for image in base:
            if image not in ["stimela/ddfacet", "radioastro/ddfacet"]:
                docker.pull(image)
                log.log_image(image)

    log.write()


def images(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all stimela related images.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
            help="Clear the logfile that keeps track of stimela images. This does not do anythig to the images.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('images')

    if args.clear:
        log.clear('images')
        log.write()


def containers(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all active stimela containers.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
            help="Clear the log file that keeps track of stimela containers. This doesn't do anything to the containers.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('containers')
    if args.clear:
        log.clear('containers')
        log.write()


def ps(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all running stimela processes')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
            help="Clear logfile that keeps track of stimela processes. This doesn't do anything ot the processes themselves.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('processes')
    if args.clear:
        log.clear('processes')
        log.write()


def kill(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Gracefully kill stimela process(s).')

    add = parser.add_argument

    add("pid", nargs="*",
            help="Process ID")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)

    for pid in args.pid:

        found = pid in log.info['processes'].keys()

        if not found:
            print("Could not find process {0}".format(pid))
            continue

        try:
            os.kill(int(pid), signal.SIGINT)
        except OSError:
            raise OSError('Process with PID {} could not be killed'.format(pid))

        log.remove('processes', pid)
    log.write()


def clean(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Convience tools for cleaning up after stimela')
    add = parser.add_argument

    add("-ai", "--all-images", action="store_true",
        help="Remove all images pulled/built by stimela. This include CAB images")
    
    add("-ab", "--all-base", action="store_true",
        help="Remove all base images")

    add("-ac", "--all-cabs", action="store_true", 
        help="Remove all CAB images")

    add("-aC", "--all-containers", action="store_true",
        help="Stop and/or Remove all stimela containers")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)

    if args.all_images:
        images = log.info['images'].keys()
        for image in images:
            utils.xrun('docker', ['rmi', image])
            log.remove('images', image)
            log.write()

    if args.all_base:
        images = log.info['images'].keys()
        for image in images:
            if log.info['images'][image]['CAB'] is False:
                utils.xrun('docker', ['rmi', image])
                log.remove('images', image)
                log.write()

    if args.all_cabs:
        images = log.info['images'].keys()
        for image in images:
            if log.info['images'][image]['CAB']:
                utils.xrun('docker', ['rmi', image])
                log.remove('images', image)
                log.write()

        
    if args.all_containers:
        containers = log.info['containers'].keys()
        for container in containers:
            cont = docker.Container(log.info['containers'][container]['IMAGE'], container)
            try:
                status = cont.info()['State']['Status']
            except OSError:
                print('Could not inspect container {}. It probably doesn\'t exist, will remove it from log'.format(container))
                status = False
                log.remove('containers', container)
                log.write()
                continue

            if status == 'running':
            # Kill the container instead of stopping it, so that effect can be felt py parent process
                utils.xrun('docker', ['kill', container])
                cont.remove()
                log.remove('containers', container)
                log.write()
            elif status == 'exited':
                cont.remove()
                log.remove('containers', container)
                log.write()


def main(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg


    parser = ArgumentParser(description='Stimela: Dockerized Radio Interferometric Scripting Framework. '
                                        '|n version {:s} |n Sphesihle Makhathini <sphemakh@gmail.com>'.format(__version__),
                            formatter_class=MultilineFormatter, add_help=False)

    add = parser.add_argument

    add("-h", "--help",  action="store_true",
            help="Print help message and exit")

    add("-v","--version", action='version',
            version='{:s} version {:s}'.format(parser.prog, __version__))

    add("command", nargs="*", metavar="command [options]",
            help="Stimela command to execute. For example, 'stimela help run'")

    options = []
    commands = dict(pull=pull, build=build, run=run,
                    images=images, cabs=cabs, ps=ps,
                    containers=containers, kill=kill,
                    clean=clean)

    command = "failure"

    for cmd in commands:
        if cmd in argv:
            command = cmd

            index = argv.index(cmd)
            options = argv[index+1:]
            argv = argv[:index+1]

    args = parser.parse_args(argv)

    # Command is help and no other commands following
    try:
        main_help = (args.command[0] == "help" and len(args.command) == 1)
    except IndexError:
        main_help = True

    if args.help or main_help:
        parser.print_help()

        print ("""
Run a command. These can be:

help    : Prints out a help message about other commands
build   : Build a set of stimela images
pull    : pull a stimela base images
run     : Run a stimela script
images  : List stimela images
cabs    : List active stimela containers
ps      : List running stimela scripts
kill    : Gracefully kill runing stimela process
clean   : Clean up tools for stimela

""")

        sys.exit(0)

    # Separate commands into command and arguments
    cmd, argv = args.command[0], args.command[1:]

    # If we've got past the if statement above, and help
    # is the command then assume that help on a command
    # is requested
    if cmd == "help":
        # Request help on the sub-command
        cmd, argv = argv[0], ["-h"]
    else:
        argv = options

    # Get the function to execute for the command
    try:
        _cmd = commands[cmd]
    except KeyError:
        raise KeyError("Command '{:s}' not recognized "
                         "Run : 'stimela help' for help".format(cmd))

    # Invoke the command
    _cmd(argv)



import os
import sys
import re

sys.path.append('/utils')
import utils

CONFIG = os.environ['CONFIG']
INPUT = os.environ['INPUT']
OUTPUT = os.environ['OUTPUT']
MSDIR = os.environ['MSDIR']

cab = utils.readJson(CONFIG)
params = cab['parameters']
args = []

for param in params:
    name = param['name']
    value = param['value']

    if name == 'msname':
        mslist = ' '.join(value)
        continue

    if value in [None, False]:
        continue
    # only versions > 2.0 have this 
    if name in ["spws"]:
        continue

    if name == 'scale':
        if isinstance(value, (int, float)):
            value = '{0}asec'.format(value)

    if name in 'size trim nwlayers-for-size beamshape channelrange interval'.split():
        if isinstance(value, int):
            value = '{0} {0}'.format(value)
        elif getattr(value, '__iter__'):
            if len(value) == 1:
                value.append( value[0])
            value = ' '.join(map(str, value))

    if name in 'spws multiscale-scales pol'.split():
        if getattr(value, '__iter__'):
            value = ','.join(map(str, value))

    if value is True:
        arg = '{0}{1}'.format(cab['prefix'], name)
    else:
        arg = '{0}{1} {2}'.format(cab['prefix'], name, value)

    args.append(arg)

utils.xrun(cab['binary'], args + [mslist])



### Don't run this unless you preparing a new release
import sys
import os
import yaml
import json
import codecs
from collections import OrderedDict

version = sys.argv[1]
root = sys.argv[2]

with open('{}/stimela_misc/version.py'.format(root), 'w') as std:
    std.write('version = \'{}\''.format(version))
    

CABS = {}
for cab in os.listdir('{}/stimela/cargo/cab'.format(root)):
    path = '{0}/stimela/cargo/cab/{1}'.format(root, cab)
    if os.path.exists('{}/parameters.json'.format(path)) \
        and os.path.exists('{}/Dockerfile'.format(path)):

        with open('{}/parameters.json'.format(path)) as _std:
            params = yaml.safe_load(_std)

        with open('{}/Dockerfile'.format(path)) as std:
            lines = std.readlines()
            changed = False
            for i,line in enumerate(lines):
                if line.startswith('FROM'):
                    new = 'FROM {0}:{1}\n'.format(params['base'], version)
                    lines[i] = new
                    changed = True

        with open('{}/Dockerfile'.format(path), 'w') as std:
            std.write( ''.join(lines))

        with codecs.open('{}/parameters.json'.format(path), 'w', 'utf8') as std:

            _params = OrderedDict([('task', params['task']),
                                  ('base', params['base']),
                                  ('tag', version),
                                  ('description', params['description']),
                                  ('prefix', params['prefix']),
                                  ('binary', params['binary']),
                                  ('msdir', params['msdir']),
                                  ('parameters', params['parameters'])])

            std.write(json.dumps(_params, indent=4, ensure_ascii=False))


for base in os.listdir('{0}/stimela/cargo/base'.format(root)):
    path = '{0}/stimela/cargo/base/{1}'.format(root, base)
    if not os.path.exists('{}/Dockerfile'.format(path)):
        continue

    with open('{}/Dockerfile'.format(path)) as std:
        lines = std.readlines()
        changed = False

        for i,line in enumerate(lines):
            if line.startswith('FROM') and line.split()[1].startswith('stimela'):
                image = line.split()[1].split(':')[0]
                new = 'FROM {0}:{1}\n'.format(image, version)
                print new
                lines[i] = new
                changed = True

    if changed:
        with open('{}/Dockerfile'.format(path), 'w') as std:
            std.write(''.join(lines))



import pyfits as fitsio
import numpy
from pyrap.tables import table
import numpy
import os
import sys 
import tempfile

sys.path.append('/utils')
import utils

imagename = sys.argv[1]
msname = sys.argv[2]
cell = sys.argv[3]

hdulist = fitsio.open(imagename)
if isinstance(hdulist, list):
    hdu = hdulist[0]
else:
    hdu = hdulist

tfile = tempfile.NamedTemporaryFile(suffix='.fits')
tfile.flush()

nchan = hdu.data.shape[0]
sys.stdout.write('Creating template image\n')
os.system('rm -fr {0}'.format(tfile.name))
utils.xrun('lwimager', ['ms='+msname, 'fits='+tfile.name, 'data=psf', 'npix=64',
                        'mode=mfs', 'nchan={:d}'.format(nchan), 'img_nchan={:d}'.format(nchan),
                        'prefervelocity=False', 'cellsize={0}'.format(cell)])

sys.stdout.write('Created template image. Will now proceed to simulate FITS model into MS\n')


with fitsio.open(tfile.name) as hdu0:
    header = hdu0[0].header

tfile.close()

naxis = hdu.header['NAXIS']
raNpix = hdu.header['NAXIS1']
decNpix = hdu.header['NAXIS2']

header['CRPIX1'] = raNpix/2.0
header['CRPIX2'] = decNpix/2.0

if naxis<2:
    raise RuntimeError('FITS image has to have at least two axes')

elif naxis == 2:
    hdu.data = hdu.data[numpy.newaxis,numpy.newaxis,...]
    sys.stdout.write('WARNING::FITS image has 2 axes. Will add STOKES and FREQ axes\n')
elif naxis == 3:
    sys.stdout.write('WARNING::FITS image has 3 axes.\n')
    # Try to find out what is the 3rd axis
    third_axis = hdu.header.get('CTYPE3', None)
    if third_axis is None:
        sys.stdout.write('WARNING:: Third axis is not labelled, assuming its the FREQ Axis\n')
        hdu.data = hdu.data[:,numpy.newaxis,...]
    elif third_axis.lower() == 'freq':
        hdu.data = hdu.data[:,numpy.newaxis,...]
    elif third_axis.lower() == 'stokes':
        hdu.data = hdu.data[numpy.newaxis,...]
    else:
        sys.stdout.write('WARNING:: CTYPE3 value [{}] is unknown, will ignore it and assume 3rd axis is FREQ\n'.format(third_axis))
        hdu.data = hdu.data[:,numpy.newaxis,...]

elif naxis == 4:
    sys.stdout.write('FITS image has four axes. If the axes are labeled (via CTYPE<axis index>), will attempt to restructure it so it is fit to predict from\n')
    third_axis = hdu.header.get('CTYPE3', None)
    if third_axis is None:
        raise RuntimeError('ABORT:: FITS axes are not labelled. Cannot proceed. Please label STOKES and FREQ accordingly\n')
    elif third_axis.lower()=='freq':
        hdu.data = numpy.rollaxis(hdu.data, 1)
else:
    sys.stdout.write('WARNING:: FITS image has more than 4 axes. Passing the  buck to lwimger\n')


hdu.header = header

hdulist.writeto(sys.argv[4], clobber=True)
